use crate::components::websocket::WebSocketContext;
use leptos::{ev, prelude::*};
use leptos_use::core::ConnectionReadyState;
use lib::*;
use serde_json::to_string;
use thaw::*;

#[component]
pub fn Auth() -> impl IntoView {
    let websocket = expect_context::<WebSocketContext>();
    let username = RwSignal::new("".to_string());
    let user_context = expect_context::<ReadSignal<Option<UserUpdate>>>();
    let connected = move || websocket.ready_state.get() == ConnectionReadyState::Open;
    let new_username = RwSignal::new(String::new());

    Effect::new(move |_| {
        user_context.with(|new_user| {
            if let Some(user) = new_user {
                username.set(user.username.to_string());
            }
        });
    });

    let send_login = move |_| {
        websocket.send(
            &to_string(&UserLogInRequest {
                username: new_username(),
            })
            .unwrap(),
        );
        new_username.set(String::new());
    };

    // Clear user name on disconnect
    Effect::new(move |_| {
        if !connected() {
            username.set(String::from(""));
        }
    });

    view! {
        <Popover trigger_type=PopoverTriggerType::Click position=PopoverPosition::BottomEnd>
            <PopoverTrigger slot>
                <Button appearance=ButtonAppearance::Transparent size=ButtonSize::Small>
                    <p class="mx-2">{move || username()}</p>
                    <Avatar class="drop-shadow-md" name=username />
                </Button>
            </PopoverTrigger>
            <Show when=move || { connected() } fallback=|| view! { <p>"Disconnected."</p> }>
                <Popover
                    trigger_type=PopoverTriggerType::Click
                    position=PopoverPosition::BottomStart
                >
                    <PopoverTrigger slot>
                        <h2 class="text-xl">
                            "Change Name"
                            <Button
                                icon=icondata::TbHelp
                                appearance=ButtonAppearance::Transparent
                                size=ButtonSize::Small
                            />
                        </h2>
                    </PopoverTrigger>
                    <p class="indent-1 text-pretty">
                        "You were already given a random name but if you'd like to change it this is the place."
                    </p>
                    <p class="my-2 indent-1 text-pretty">
                        "Identities are saved once created so if you leave or get disconnected just enter your old name here to \"log back in\"."
                    </p>
                    <p class="indent-1 text-pretty">
                        "Please don't steal each other's identities (yes, you can)."
                    </p>
                </Popover>
                <br />

                <form onsubmit="return false" on:submit=send_login.clone() autocomplete="off">
                    <FieldContextProvider>
                        <Field label="Username" name="username">
                            <Input
                                placeholder=username
                                value=new_username
                                disabled=!connected()
                                class="w-80"
                                rules=vec![InputRule::required(true.into())]
                            >
                                <InputPrefix slot>
                                    <Icon icon=icondata::AiUserOutlined />
                                </InputPrefix>
                            </Input>
                        </Field>
                        <Button
                            button_type=ButtonType::Submit
                            on_click={
                                let field_context = FieldContextInjection::expect_context();
                                move |e: ev::MouseEvent| {
                                    if !field_context.validate() {
                                        e.prevent_default();
                                    }
                                }
                            }
                        >
                            "Submit"
                        </Button>
                    </FieldContextProvider>
                </form>
            </Show>
        </Popover>
    }
}
