use crate::components::websocket::WebSocketContext;
use leptos::{html::Textarea, prelude::*};
use leptos_use::core::ConnectionReadyState;
use lib::*;
use serde_json::to_string;
use thaw::*;

#[component]
pub fn Chat() -> impl IntoView {
    let websocket = expect_context::<WebSocketContext>();
    let chat_context = expect_context::<ReadSignal<Option<ChatMessage>>>();
    let chat_update_context = expect_context::<ReadSignal<Option<ChatUpdate>>>();

    // Chat stuff
    let chat_history = RwSignal::<Vec<String>>::new(vec![]);
    let users = RwSignal::<Vec<String>>::new(vec![]);
    let chat_name = RwSignal::<String>::new("".to_string());
    let chat_input = RwSignal::new(String::new());
    let chat_history_ref = NodeRef::<Textarea>::new();

    // Connection status updates in chat window
    Effect::new(move |_| {
        websocket.ready_state.with(move |state| match *state {
            ConnectionReadyState::Connecting => {
                chat_history
                    .update(|history| history.push("Connecting to game server...\n".to_string()));
            }
            ConnectionReadyState::Open => {
                chat_history.update(|history| history.push("Connected!\n".to_string()));
            }
            ConnectionReadyState::Closing => {
                chat_history.update(|history| history.push("Disconnecting...\n".to_string()));
            }
            ConnectionReadyState::Closed => {
                chat_history.update(|history| history.push("Disconnected.\n".to_string()));
                users.update(|users| users.clear());
            }
        })
    });

    // Handle incoming chat messages
    Effect::new(move |_| {
        chat_context.with(move |chat_message| {
            if let Some(message) = chat_message {
                chat_history.update(|history| history.push(format!("{}\n", message.text)));
            }
        })
    });

    // Handle users
    Effect::new(move |_| {
        chat_update_context.with(move |chat_update| {
            if let Some(update) = chat_update {
                users.set(update.users.clone());
                chat_name.set(update.room.clone());
            }
        })
    });

    // Handle sending messages
    let send_message = move |_| {
        if chat_input() != "" {
            websocket.send(&to_string(&ChatMessage { text: chat_input() }).unwrap());
            chat_input.set(String::new());
        }
    };

    // Keep chat scrolled to the bottom
    Effect::new(move |_| {
        chat_history.with(move |_| {
            // Scroll chat textarea to bottom
            if let Some(hist) = chat_history_ref.get() {
                hist.set_scroll_top(hist.scroll_height());
            }
        })
    });

    view! {
        <div class="lg:flex lg:justify-between">
            <div class="w-full">
                <h2 class="text-xl">Chat: {move || chat_name()}</h2>
                <textarea
                    node_ref=chat_history_ref
                    class="opacity-80 w-full h-48 rounded-md resize-none bg-neutral-900 text-neutral-200 drop-shadow-md text-sm"
                    readonly=true
                    wrap="soft"
                >
                    {move || chat_history.get()}
                </textarea>
                <br />
                <form onsubmit="return false" on:submit=send_message autocomplete="off">
                    <Input
                        class="drop-shadow-md w-full"
                        value=chat_input
                        placeholder="talk shit..."
                    />
                    <Button button_type=ButtonType::Submit class="drop-shadow-md">
                        "Send"
                    </Button>
                </form>
            </div>
            <ul class="mx-2">
                <h2 class="text-xl">Users: {move || users().len()}</h2>
                {move || {
                    users()
                        .into_iter()
                        .map(|n| {
                            view! {
                                <li>
                                    <Avatar name=n.clone() />
                                    {n}
                                </li>
                            }
                        })
                        .collect_view()
                }}
            </ul>
        </div>
    }
}
