// use bitcode::decode;
use codee::string::FromToStringCodec;
use leptos::prelude::*;
use leptos_use::{core::ConnectionReadyState, use_websocket, UseWebSocketReturn};
use lib::*;
use serde_json::from_str;
use std::sync::Arc;

#[derive(Clone)]
pub struct WebSocketContext {
    pub ready_state: Signal<ConnectionReadyState>,
    // pub message: Signal<Option<String>>,
    send: Arc<dyn Fn(&String) + Send + Sync + 'static>,
    pub open: Arc<dyn Fn() + Send + Sync + 'static>,
    pub close: Arc<dyn Fn() + Send + Sync + 'static>,
}

impl WebSocketContext {
    pub fn new(
        ready_state: Signal<ConnectionReadyState>,
        // message: Signal<Option<String>>,
        send: Arc<dyn Fn(&String) + Send + Sync + 'static>,
        open: Arc<dyn Fn() + Send + Sync + 'static>,
        close: Arc<dyn Fn() + Send + Sync + 'static>,
    ) -> Self {
        Self {
            ready_state,
            // message,
            send,
            open,
            close,
        }
    }

    #[inline(always)]
    pub fn send(&self, message: &str) {
        (self.send)(&message.to_string())
    }
}

#[component]
pub fn Websocket() -> impl IntoView {
    let url;

    if cfg!(debug_assertions) {
        url = "ws://127.0.0.1:3030/websocket";
    } else {
        url = "wss://cards.doordesk.net/websocket";
    }

    let UseWebSocketReturn {
        ready_state,
        message,
        send,
        open,
        close,
        ..
    } = use_websocket::<String, String, FromToStringCodec>(url);

    provide_context(WebSocketContext::new(
        ready_state,
        // message,
        Arc::new(send.clone()),
        Arc::new(open.clone()),
        Arc::new(close.clone()),
    ));

    let connected = move || ready_state.get() == ConnectionReadyState::Open;

    // Contexts for message handler
    // TODO: This context stuff can probably be done better
    let (users_count, set_users_count) = signal::<Option<ServerOnlineUsers>>(Option::None);
    let (games_count, set_games_count) = signal::<Option<ServerActiveGames>>(Option::None);
    let (active_games, set_active_games) = signal::<Vec<GameBrowserMeta>>(vec![]);
    let (user_update, set_user_update) = signal::<Option<UserUpdate>>(Option::None);
    let (chat_update, set_chat_update) = signal::<Option<ChatUpdate>>(Option::None);
    let (judge_round, set_judge_round) = signal::<Option<JudgeRound>>(Option::None);
    let (chat_message, set_chat_message) = signal::<Option<ChatMessage>>(Option::None);
    let (game_meta, set_game_meta) = signal::<Option<GameMeta>>(Option::None);
    let (game_state, set_game_state) = signal::<Option<GameStateMeta>>(Option::None);
    let (card_packs_meta, set_card_packs_meta) = signal::<CardPacksMeta>(CardPacksMeta {
        official_meta: vec![],
        unofficial_meta: vec![],
    });

    Effect::new(move |_| {
        if !connected() {
            set_active_games(vec![]);
            set_user_update(None);
            set_chat_update(None);
            set_judge_round(None);
            set_chat_message(None);
            set_game_meta(None);
            set_game_state(None);
        }
    });

    provide_context::<ReadSignal<CardPacksMeta>>(card_packs_meta);
    provide_context::<ReadSignal<Option<UserUpdate>>>(user_update);
    provide_context::<ReadSignal<Option<ChatUpdate>>>(chat_update);
    provide_context::<ReadSignal<Option<JudgeRound>>>(judge_round);
    provide_context::<WriteSignal<Option<JudgeRound>>>(set_judge_round);
    provide_context::<ReadSignal<Option<ChatMessage>>>(chat_message);
    provide_context::<ReadSignal<Vec<GameBrowserMeta>>>(active_games);
    provide_context::<ReadSignal<Option<GameMeta>>>(game_meta);
    provide_context::<ReadSignal<Option<GameStateMeta>>>(game_state);
    provide_context::<ReadSignal<Option<ServerOnlineUsers>>>(users_count);
    provide_context::<ReadSignal<Option<ServerActiveGames>>>(games_count);

    // Incoming Message handler
    Effect::new(move |_| {
        message.with(move |message_raw| {
            if let Some(message) = message_raw {
                if let Ok(message) = from_str::<ServerToClientMessage>(message) {
                    match message {
                        ServerToClientMessage::CardPacksMeta(card_packs_meta) => {
                            set_card_packs_meta(card_packs_meta);
                        }

                        ServerToClientMessage::ChatMessage(chat_message) => {
                            set_chat_message(Some(chat_message));
                        }
                        ServerToClientMessage::ChatUpdate(chat_update) => {
                            set_chat_update(Some(chat_update));
                        }

                        ServerToClientMessage::GameMeta(game_meta) => {
                            set_game_meta(Some(game_meta));
                        }

                        ServerToClientMessage::GameStateMeta(game_state_meta) => {
                            set_game_state(Some(game_state_meta));
                        }

                        ServerToClientMessage::GamesUpdate(games_update) => {
                            set_active_games(games_update.games);
                        }

                        ServerToClientMessage::JudgeRound(judge_round) => {
                            set_judge_round(Some(judge_round));
                        }

                        ServerToClientMessage::ServerActiveGames(server_active_games) => {
                            set_games_count(Some(server_active_games));
                        }

                        ServerToClientMessage::ServerOnlineUsers(server_online_users) => {
                            set_users_count(Some(server_online_users));
                        }

                        ServerToClientMessage::UserUpdate(user_update) => {
                            set_user_update(Some(user_update));
                        }
                    }
                } else {
                    leptos::logging::log!("Unhandled message: {:#?}", message);
                }
            }
        });
    });
}
