use crate::user_handler::*;
use crate::AppState;
use crate::GameHandlerMessage::*;
use crate::UserHandlerMessage::*;
use axum::extract::ws::{CloseFrame, Message};
use lib::*;
use serde_json::from_str;
use std::{net::SocketAddr, sync::Arc};

/// Handles incoming messages
pub struct IncomingMessageHandler {
    /// Global state pointer
    state: Arc<AppState>,
}

impl IncomingMessageHandler {
    /// Returns new IncomingMessageHandler object
    pub fn new(state: Arc<AppState>) -> Self {
        IncomingMessageHandler { state }
    }

    /// Handles incoming messages
    pub async fn handle(&self, addr: SocketAddr, message: Message) {
        match message {
            Message::Text(text) => match text {
                _chat_message if let Ok(chat_message) = from_str::<ChatMessage>(&text) => {
                    // TODO: This should be delegated to user handler and an outgoing message and/or chat handler
                    // let msg;
                    if chat_message.text.len() > 1024 {
                        if let Err(e) = self
                            .state
                            .tx_user_handler
                            .send(PassChatMessage(chat_message.text[..1024].to_string(), addr))
                            .await
                        {
                            tracing::error!("Error contacting outgoing message handler: {}", e);
                        }
                    } else {
                        if let Err(e) = self
                            .state
                            .tx_user_handler
                            .send(PassChatMessage(chat_message.text, addr))
                            .await
                        {
                            tracing::error!("Error contacting outgoing message handler: {}", e);
                        }
                    }
                }
                _user_log_in_request
                    if let Ok(user_log_in) = from_str::<UserLogInRequest>(&text) =>
                {
                    if let Err(e) = self
                        .state
                        .tx_user_handler
                        .send(UserLogIn(user_log_in, addr))
                        .await
                    {
                        tracing::error!("Error sending user login: {}", e)
                    }
                }
                _new_game_request if let Ok(new_game) = from_str::<NewGameRequest>(&text) => {
                    if let Err(e) = self
                        .state
                        .tx_user_handler
                        .send(GetUser(new_game, addr))
                        .await
                    {
                        tracing::error!("Error requesting new game: {}", e)
                    }
                }

                _join_game_request if let Ok(join_request) = from_str::<GameJoinRequest>(&text) => {
                    if let Err(e) = self
                        .state
                        .tx_user_handler
                        .send(UserHandlerMessage::JoinGame(join_request.id, addr))
                        .await
                    {
                        tracing::error!("Error requesting game join: {}", e)
                    }
                }

                _player_move_request
                    if let Ok(move_request) = from_str::<PlayerMoveRequest>(&text) =>
                {
                    if move_request.card_ids.is_empty() {
                        tracing::error!("Move request card_ids is empty! Ignoring...");
                        return;
                    }
                    if move_request.game_id == "".to_string() {
                        tracing::error!("Move request game_id is empty! Ignoring...");
                        return;
                    } else {
                        if let Err(e) = self
                            .state
                            .tx_user_handler
                            .send(UserHandlerMessage::MoveRequest(move_request, addr))
                            .await
                        {
                            tracing::error!("Error sending move request: {}", e)
                        }
                    }
                }

                _judge_decision
                    if let Ok(judge_decision) = from_str::<JudgeDecisionRequest>(&text) =>
                {
                    if !judge_decision.winning_cards.is_empty() {
                        if let Err(e) = self
                            .state
                            .tx_user_handler
                            .send(GetPlayerUserId(judge_decision, addr))
                            .await
                        {
                            tracing::error!("Error contacting user handler: {}", e)
                        }
                    } else {
                        tracing::error!("Judge request received with empty cards");
                    }
                }

                _delete_game if let Ok(delete_request) = from_str::<GameDeleteRequest>(&text) => {
                    if let Err(e) = self
                        .state
                        .tx_game_handler
                        .send(DeleteGame(delete_request))
                        .await
                    {
                        tracing::error!("Error sending delete game: {}", e)
                    }
                }

                _ => tracing::error!(
                    "Unhandled text from {}\n{:#?}\nTODO: Probably close connection here",
                    addr,
                    &text
                ),
            },

            Message::Binary(data) => tracing::error!(
                "{} sent binary: {:?}\n TODO: Probably close connection here",
                addr,
                data
            ),

            Message::Close(close_frame) => self.handle_close(close_frame, addr).await,

            Message::Ping(ping) => {
                tracing::info!("Pong received with: {:?}", ping);
            }

            Message::Pong(pong) => {
                tracing::info!("Pong received with: {:?}", pong);
            }
        }
    }

    /// This runs when a connection closes
    async fn handle_close(&self, close_frame: Option<CloseFrame>, addr: SocketAddr) {
        // Process close frame
        if let Some(cf) = close_frame {
            tracing::info!(
                "Close received with code: {} and reason: {}",
                cf.code,
                cf.reason
            )
        } else {
            tracing::info!("close received without close frame")
        }

        if let Err(e) = self
            .state
            .tx_user_handler
            .send(UserHandlerMessage::Cleanup(addr))
            .await
        {
            tracing::info!("Error contacting user handler: {}", e)
        }
    }
}
