use crate::game_handler::*;
use crate::incoming_message_handler::*;
use crate::outgoing_message_handler::*;
use crate::websocket::*;
use anyhow::{Context, Result};
use axum::{routing::get, Router};
use clap::{arg, command};
use server::*;
use std::{net::SocketAddr, sync::Arc, time::Duration};
use tokio::sync::{broadcast, mpsc};
use tower::ServiceBuilder;
use tower_governor::{governor::GovernorConfigBuilder, GovernorLayer};
use tower_http::{compression::CompressionLayer, services::ServeDir};
use tracing_subscriber::prelude::*;
use user_handler::UserHandler;

// #[tokio::main(flavor = "current_thread")]
#[tokio::main]
async fn main() -> Result<()> {
    // Handle command-line args
    let matches = command!()
        .arg(
            arg!(-b --bind <ADDR> "Bind to specific address. Default is 127.0.0.1:3030")
                .required(false),
        )
        .arg(
            arg!(-v --verbose "Extra verbosity for use with tokio-console or to satisfy curiosity")
                .required(false),
        )
        .get_matches();

    let bind_addr;
    if let Some(addr) = matches.get_one::<String>("bind") {
        bind_addr = String::from(addr);
    } else {
        bind_addr = String::from("127.0.0.1:3030");
    }

    let trace_level;
    if *matches.get_one::<bool>("verbose").unwrap() {
        trace_level = "server=trace,tower_http=trace,lib=trace,tokio=trace,runtime=trace"
    } else {
        trace_level = "server=trace,tower_http=trace,lib=trace";
    }

    // Set up tracing
    tracing_subscriber::registry()
        .with(
            tracing_subscriber::EnvFilter::try_from_default_env()
                .unwrap_or_else(|_| trace_level.into()),
        )
        .with(console_subscriber::ConsoleLayer::builder().spawn())
        .with(tracing_subscriber::fmt::layer())
        .init();

    // Set up rate limiting/governor
    let governor_conf = Arc::new(
        GovernorConfigBuilder::default()
            .per_second(20)
            .burst_size(200)
            .finish()
            .unwrap(),
    );

    let interval = Duration::from_secs(60);
    let governor_limiter = governor_conf.limiter().clone();

    std::thread::spawn(move || loop {
        std::thread::sleep(interval);
        tracing::info!("rate limiting storage size: {}", governor_limiter.len());
        governor_limiter.retain_recent();
    });

    // Set up globals
    let (tx_broadcast, _rx_broadcast) = broadcast::channel(32);
    let (tx_game_handler, mut rx_game_handler) = mpsc::channel(32);
    let (tx_incoming_message_handler, mut rx_incoming_message_handler) = mpsc::channel(32);
    let (tx_outgoing_message_handler, mut rx_outgoing_message_handler) = mpsc::channel(32);
    let (tx_user_handler, mut rx_user_handler) = mpsc::channel(32);

    let app_state = Arc::new(AppState {
        tx_broadcast: tx_broadcast.clone(),
        tx_game_handler,
        tx_incoming_message_handler,
        tx_outgoing_message_handler,
        tx_user_handler,
    });

    // Spawn task to handle incoming messages
    let incoming_message_handler = IncomingMessageHandler::new(app_state.clone());
    tokio::task::Builder::new()
        .name("Incoming Message Handler")
        .spawn(async move {
            while let Some((addr, message)) = rx_incoming_message_handler.recv().await {
                incoming_message_handler.handle(addr, message).await
            }
        })
        .unwrap();

    // Spawn task to handle outgoing messages
    let outgoing_message_handler = OutgoingMessageHandler::new(tx_broadcast);
    tokio::task::Builder::new()
        .name("Outgoing Message Handler")
        .spawn(async move {
            while let Some(message) = rx_outgoing_message_handler.recv().await {
                outgoing_message_handler.handle(message).await
            }
        })
        .unwrap();

    // Spawn task to handle User things
    let mut user_handler = UserHandler::new(app_state.clone());
    tokio::task::Builder::new()
        .name("User Handler")
        .spawn(async move {
            while let Some(message) = rx_user_handler.recv().await {
                user_handler.handle(message).await
            }
        })
        .unwrap();

    // Spawn task to handle Game things
    let mut game_handler = GameHandler::new(app_state.clone());
    tokio::task::Builder::new()
        .name("Game Handler")
        .spawn(async move {
            while let Some(message) = rx_game_handler.recv().await {
                game_handler.handle(message).await
            }
        })
        .unwrap();

    // Router
    let app = Router::new()
        .route("/websocket", get(websocket_connection_handler))
        .fallback_service(ServeDir::new("dist"))
        .layer(ServiceBuilder::new().layer(CompressionLayer::new()))
        .layer(GovernorLayer {
            config: governor_conf,
        })
        .with_state(app_state);

    // send it
    let listener = tokio::net::TcpListener::bind(bind_addr.to_owned())
        .await
        .with_context(|| {
            format!(
                "{} is not a valid socket address. Did you forget to specify a port?",
                bind_addr
            )
        })?;
    tracing::info!("listening on {}", listener.local_addr()?);
    axum::serve(
        listener,
        app.into_make_service_with_connect_info::<SocketAddr>(),
    )
    .await?;

    Ok(())
}
